import _bs
from _bs import *

# private stuff doesnt come in by default but we have lots of code assuming this stuff lives in bs...
from _bs import _consolePrint,_startListeningForWiiRemotes,_stopListeningForWiiRemotes,_setIsPlaying,\
    _setTouchscreenEditing,_setHaveMods,_debugPrintPyErr,_focusWindow,_isLogFull,_getDemoTimeRemaining,\
    _purchaseGame,_setInternalLanguageKeys,_newMaterial,_materialAddAction,_submitScore,_storeRemoteConfig,\
    _requestRemoteConfigKeys,_getLog,_getUUID,_setDemoTime,_markLogSent,_reportAchievement

import bs
import os
import time
from bsVector import Vector
import random
import weakref
import ast
import gc
import types

_gLanguage = None

def getGameTypes():
    allGames = []
    haveMods = False
    for i,d in enumerate([bs.getSystemScriptsDirectory(),bs.getUserScriptsDirectory()]):
        for name in os.listdir(d):
            try:
                if name == 'sys' or name.endswith('.py'):
                    # if there's anything in user-mods, let the game know that we've got mods going on
                    # (we do less error reporting in this case for my sanity's sake)
                    if i == 1: haveMods = True
                    
                if name.endswith('.py'):
                    module = __import__(name.replace('.py',''))
                    if hasattr(module,'getGames'):
                        games = module.getGames()
                        allGames += games
            except Exception,e:
                print 'Error importing game module \''+name+'\':',e

    bs._setHaveMods(haveMods)
    return allGames


class WeakCall(object):
    """
    Wraps a callable and arguments into a single callable object.
    If the callable is a bound method, it is weak-referenced, meaning
    the underlying object is free to die if not referenced from elsewhere.
    Should this occur, calling this object is simply a no-op.
    """
    def __init__(self,*args,**keywds):
        if hasattr(args[0],'im_func'): self._call = WeakMethod(args[0])
        else: self._call = args[0]
        self._args = args[1:]
        self._keywds = keywds
    def __call__(self,*argsExtra):
        return self._call(*self._args+argsExtra,**self._keywds)

class Call(object):
    """
    Wraps a callable and arguments into a single callable object.
    The callable is strong-referenced so it won't die until this object does.
    """
    def __init__(self,*args,**keywds):
        self._call = args[0]
        self._args = args[1:]
        self._keywds = keywds
    def __call__(self,*argsExtra):
        return self._call(*self._args+argsExtra,**self._keywds)

class WeakMethod :
    """ wraps a bound method using weak references so that the original is free to die.
    If called with a dead target, is simply a no-op """
    def __init__( self , f ) :
        self.f = f.im_func
        self.c = weakref.ref( f.im_self )
    def __call__( self , *arg ) :
        c = self.c()
        if c is None : return
        return apply(self.f,(c,)+arg)


class AttrDict(dict):
    """ a dict that can be accessed with dot notation: foo.bar is equivalant to foo['bar'] """
    def __getattr__(self, attr):
        return self[attr]
    def __setattr__(self, attr, value):
        self[attr] = value

def _addToAttrDict(dst,src):
    for key,value in src.items():
        if type(value) is dict:
            try: dstDict = dst[key]
            except Exception: dstDict = dst[key] = AttrDict()
            if type(dstDict) is not AttrDict: raise Exception("language key '"+key+"' is defined both as a dict and value")
            _addToAttrDict(dstDict,value)
        else:
            if type(value) not in (float,int,bool,str,unicode,type(None)): raise Exception("invalid value type for res '"+key+"': "+str(type(value)))
            dst[key] = value

def getLanguage():
    """Returns the current language"""
    try: language = bs.config['Language']
    except Exception: language = 'English'
    return language
    
def setLanguage(language,printChange=True):
    """Set the language used by the game.  Pass None to use OS default."""
    global _gLanguage

    try: curLanguage = bs.config['Language']
    except Exception: curLanguage = None

    # store this in the config if its changing
    if language != curLanguage:
        if language is None:
            del bs.config['Language'] # clear it out for default
        else:
            bs.config['Language'] = language
        writeConfig()
        switched = True
    else: switched = False

    # None implies default (currently locked to english)
    if language is None: language = 'English'

    # create a single flattened dict for this language for simplicity
    languages = []

    try:
        l = __import__('bsLanguage'+language)
        if language != 'English': reload(l) # helpful for iterating
    except Exception,e:
        print 'Exception importing language module \'bsLanguage'+language+'\':',e
        bs.screenMessage("Error setting language to '"+language+"'; see log for details",color=(1,0,0))
        switched = False
        l = None

    if l is not None: languages.append(l)
    # everything inherits from english..
    if language != 'English': languages.append(__import__('bsLanguageEnglish'))

    languages.reverse()
    lFull = AttrDict()
    for l in languages:
        _addToAttrDict(lFull,l.values)
    _gLanguage = lFull

    # pass the root level keys/values in for low level code to use
    bs._setInternalLanguageKeys([v for v in lFull['internal'].items() if type(v[1]) in (str,unicode)])

    if switched and printChange:
        bs.screenMessage("Language is now '"+language+"'.\n(Note: this will only affect new UI/Game elements)",color=(0,1,0))

def getResource(resource,default=None):
    """returns a named resource value.  If a value is passed for 'default', it is returned in the case that no resource value is found.  Otherwise an exception is raised."""
    try:
        global _gLanguage
        if _gLanguage is None:
            language = getLanguage()
            try: setLanguage(language,printChange=False)
            except Exception,e:
                print 'Error setting language to '+language+': ',e
                # try english as a fallback
                if (language != 'English'):
                    print 'Resorting to fallback language (English)'
                    try: setLanguage('English',printChange=False)
                    except Exception,e: print 'Error setting language to English fallback: ',e

        splits = resource.split('.')
        dicts = splits[:-1]
        key = splits[-1]
        values = _gLanguage
        for d in dicts: values = values[d]
        return values[key]

    except Exception,e:
        if default is not None: return default
        else: raise Exception("resource not found: '"+resource+"'")

def translate(category,s):
    """given a category and string, returns a translated string if available; otherwise returns the original."""
    try: translated = bs.getResource('translations')[category][s]
    except Exception,e:
        print 'Translate error: cat',category,'s',s,'exc',e
        translated = None
    if translated is None: translated = s
    return translated

gTips = []
gConfigFileIsHealthy = False
gCloudConfig = {}

# config keys that we attempt to store in the cloud to share between devices
gCloudConfigKeys = ['launchCount','Campaigns','Player Profiles','Submit Debug Info','Default Player Profiles',
                    'Team Tournament Playlists','Team Tournament Playlist Randomize','Team Tournament Playlist Selection',
                    'Free-for-All Playlists','Free-for-All Playlist Randomize','Free-for-All Playlist Selection',
                    'Selected Coop Game','Show Player Names','Kick Idle Players']

# we sync achievements to/from the cloud only if we have no game-service... Otherwise that's the game-service's job
if bs.getGameService() is None:
    gCloudConfigKeys.append('Achievements')


class Player(object):

    def __init__(self,playerID,team,character,color,highlight,lobby):
        self.playerID = playerID
        self._team = weakref.ref(team)
        self.actor = None
        self.character = character
        self.color = color
        self.highlight = highlight
        self.data = {} # for custom user data
        self._lobby = weakref.ref(lobby)

    def getTeam(self):
        return self._team()

    def getIcon(self):
        return self._lobby().getPlayerIcon(self.playerID)

    def isAlive(self):
        return ( self.actor is not None and not self.actor.dead )

    def getName(self):
        return bs.getPlayerName(self.playerID)

class Team(object):
    def __init__(self,teamID=0,name='',color=(1,1,1)):
        self.teamID=teamID
        self.name=name
        self.color=color
        self.players={}
        self.data = {} # for use by activities
        self.sessionData = {} # for use by sessions

class OutOfBoundsMessage(object):
    """Tells a node that they are out of bounds"""
    def __init__(self):
        pass

class DieMessage(object):
    """Tells the Actor to die. This can happen in a delayed fashion unless immediate is true"""
    def __init__(self,immediate=False,how="generic"):
        self.immediate = immediate
        self.how = how

class StandMessage(object):
    """Tells the Actor to position itself to be standing upright on the given point"""
    def __init__(self,pos=(0,0,0),angle=0):
        self.pos = pos
        self.angle = angle

class PickUpMessage(object):
    """Tells the Actor its picked something up"""
    def __init__(self,node):
        self.node = node

class DropMessage(object):
    """Tells the Actor its dropped what it was holding"""
    def __init__(self):
        pass

class PickedUpMessage(object):
    """Tells the Actor its been picked up"""
    def __init__(self,node):
        self.node = node

class DroppedMessage(object):
    """Tells the Actor its been dropped"""
    def __init__(self,node):
        self.node = node

class ShouldShatterMessage(object):
    """Tells the Actor its hit a wall or something and should shatter"""
    def __init__(self):
        pass

class ImpactDamageMessage(object):
    """Tells the actor its been jarred violently and may want to be damaged"""
    def __init__(self,intensity):
        self.intensity = intensity

class FreezeMessage(object):
    def __init__(self):
        pass

class ThawMessage(object):
    def __init__(self):
        pass

class HitMessage(object):
    def __init__(self,srcNode=0,pos=Vector(0,0,0),velocity=Vector(0,0,0),
                 magnitude=1.0,velocityMagnitude=0.0,radius=1.0,source=None,kickBack=1.0,flatDamage=None,hitType='generic',forceDirection=None,hitSubType='default'):
        self.srcNode = srcNode
        self.pos = pos
        self.velocity = velocity
        self.magnitude = magnitude
        self.velocityMagnitude = velocityMagnitude
        self.radius = radius
        self.source = source
        self.kickBack = kickBack
        self.flatDamage=flatDamage
        self.hitType=hitType
        self.hitSubType=hitSubType
        self.forceDirection=forceDirection if forceDirection != None else velocity

class Actor(object):
    """A simple entity supporting messaging; generally consists of one or more Nodes and logic to wrangle them"""

    # RETAINED actors are kept alive by their owning activity so you don't
    # need to worry about storing references to them
    UNRETAINED = 0
    RETAINED = 1

    def __init__(self,activity,retain):
        if activity is None:
            if retain: raise Exception("actor can't be retained without an activity")
            self._activity = None
        else:
            self._activity = weakref.ref(activity)
            if retain: activity._addTransientActor(self)
            else: activity._addWeakTransientActor(self)

    def __del__(self):
        self.handleMessage(DieMessage())

    def handleMessage(self,m):
        pass

    # actors should return False here when they've died and disappeared
    # ie: at this point it is assumed the actor can be deleted without
    # visibly affecting the game
    def exists(self):
        # as a sensible default just use the existance of 'node'
        return bs.nodeExists(self.node)

    def getActivity(self):
        try: return self._activity()
        except Exception: return None

class NodeActor(Actor):
    "A simple actor consisting of a single specified node that gets deleted on the actor's death"
    def __init__(self,activity,retain,nodeType,attrs,name=None,delegate=None):
        Actor.__init__(self,activity,retain)
        owner = activity if activity is not None else bs # fixme need to disallow None as activity..
        self.node = bs.newNode(nodeType,owner=owner,attrs=attrs,name=name,delegate=delegate)
    def handleMessage(self,m):
        if isinstance(m,DieMessage):
            bs.deleteNode(self.node,quiet=True)

class Material(object):
    def __init__(self,name):
        self.materialID = bs._newMaterial()
        self.name = name
    def addAction(self,*args,**keywds):
        bs._materialAddAction(self.materialID,*args,**keywds)
    def getName(self):
        return self.name
    def __del__(self):
        pass

def _prettifyList(l,indent=0,multiLine=True):
    bits = []
    bits.append('[\n' if multiLine else '[')
    for i in range(len(l)):
        a = l[i]
        if isinstance(a,dict):
            if multiLine: bits.append(' '*(indent+3))
            bits += _prettifyDict(a,indent+3,multiLine)
        else:
            if multiLine: bits.append(' '*(indent+3) + repr(a))
            else: bits.append(repr(a))
        if (i < (len(l)-1)):
            bits.append(',')
        if multiLine: bits.append("\n")
    bits.append((' '*indent + ']') if multiLine else ']')
    return bits

def _prettifyDict(d,indent=0,multiLine=True):
    bits = []
    bits.append('{\n' if multiLine else '{')
    items = d.items()
    items.sort()
    for i in range(len(items)):
        a = items[i][0]
        b = items[i][1]
        if isinstance(b,dict):
            if multiLine: bits.append(' '*(indent+3) + repr(a) + ':')
            else: bits.append(repr(a) + ':')
            bits += _prettifyDict(b,indent+3,multiLine)
        elif isinstance(b,list):
            if multiLine: bits.append(' '*(indent+3) + repr(a) + ':')
            else: bits.append(repr(a) + ':')
            bits += _prettifyList(b,indent+3,multiLine)
        else:
            if multiLine: bits.append(' '*(indent+3) + repr(a) + ':' + repr(b))
            else: bits.append(repr(a) + ':' + repr(b))
        if (i < (len(items)-1)):
            bits.append(',')
        if multiLine: bits.append("\n")
    bits.append((' '*indent + '}') if multiLine else '}')
    return bits

def getNextTip():
   """ returns the next tip to be displayed """
   global gTips
   if len(gTips) == 0:
       for tip in getAllTips():
           gTips.insert(random.randint(0,len(gTips)),tip)
   return gTips.pop()

def getAllTips():
    tips =  [
        'If you are short on controllers, install the \'BombSquad Remote\' app\non your iOS or Android devices to use them as controllers.',
        'Create player profiles for yourself and your friends with\nyour preferred names and appearances instead of using random ones.',
        'You can \'aim\' your punches by spinning left or right.\nThis is useful for knocking bad guys off edges or scoring in hockey.',
        'If you pick up a curse, your only hope for survival is to\nfind a health powerup in the next few seconds.',
        'A perfectly timed running-jumping-spin-punch can kill in a single hit\nand earn you lifelong respect from your friends.',
        'Always remember to floss.',
        'Don\'t run all the time.  Really.  You will fall off cliffs.',
        'In Capture-the-Flag, your own flag must be at your base to score, If the other\nteam is about to score, stealing their flag can be a good way to stop them.',
        'If you get a sticky-bomb stuck to you, jump around and spin in circles. You might\nshake the bomb off, or if nothing else your last moments will be entertaining.',
        'You don\'t need to edit your profile to change characters; Just press the top\nbutton (pick-up) when joining a game to override your default.',
        'You take damage when you whack your head on things,\nso try to not whack your head on things.',
        'If you kill an enemy in one hit you get double points for it.',
        'Despite their looks, all characters\' abilities are identical,\nso just pick whichever one you most closely resemble.',
        'You can throw bombs higher if you jump just before throwing.',
        'Throw strength is based on the direction you are holding.\nTo toss something gently in front of you, don\'t hold any direction.',
        'If someone picks you up, punch them and they\'ll let go.\nThis works in real life too.',
        'Don\'t get too cocky with that energy shield; you can still get yourself thrown off a cliff.',
        'Many things can be picked up and thrown, including other players.  Tossing\nyour enemies off cliffs can be an effective and emotionally fulfilling strategy.',
        'Ice bombs are not very powerful, but they freeze\nwhoever they hit, leaving them vulnerable to shattering.',
        'Run back and forth before throwing a bomb\nto \'whiplash\' it and throw it farther.',
        'Punches do more damage the faster your fists are moving,\nso try running, jumping, and spinning like crazy.',
        'In hockey, you\'ll maintain more speed if you turn gradually.',
        'The head is the most vulnerable area, so a sticky-bomb\nto the noggin usually means game-over.',
        'Hold down any button to run. You\'ll get places faster\nbut won\'t turn very well, so watch out for cliffs.',
        'Powerups only have time limits in co-op games.\nIn teams and free-for-all they\'re yours until you die.',
        'You can judge when a bomb is going to explode based on the\ncolor of sparks from its fuse:  yellow..orange..red..BOOM.',
        ]

    ua = bs.getUserAgentString()

    # graphics settings applies to desktops and android
    if bs.getInterfaceType() == 'desktop' or ('android' in ua and not 'OUYA' in ua):
        tips += [
            'If your framerate is choppy, try turning down resolution\nor visuals in the game\'s graphics settings.',
            ]
    ua = bs.getUserAgentString()

    # auto-kick players applies to desktop and ouya
    if bs.getInterfaceType() == 'desktop' or 'OUYA' in ua:
        tips += [
            'If you\'ve got lots of players coming and going, turn on \'auto-kick-idle-players\'\nunder settings in case anyone forgets to leave the game.',
            ]

    # hot-plugging is currently mac/android only
    if 'Mac' in ua or 'android' in ua:
        tips += [
            'Players can join and leave in the middle of most games,\nand you can also plug and unplug gamepads on the fly.',
        ]

    return tips

def readConfig():
    configFilePath = bs.getConfigFilePath()
    try:
        if os.path.exists(configFilePath):
            bs.config = eval(open(configFilePath).read())
        else: bs.config = {}
        global gConfigFileIsHealthy
        gConfigFileIsHealthy = True
    except Exception,e:
        print 'error reading config file: \''+configFilePath+'\':\n',e
        # whenever this happens lets back up the broken one just in case it gets overwritten accidentally
        print 'backing up current config file to \''+configFilePath+".broken\'"
        try:
            import shutil
            shutil.copyfile(configFilePath,configFilePath+'.broken')
        except Exception,e: print 'EXC copying broken config:',e
        try: bs.log('broken config contents:\n'+open(configFilePath).read().replace('\000','<NULL_BYTE>'),toConsole=False)
        except Exception,e: print 'EXC logging broken config contents:',e
        bs.config = {}

    global gCloudConfigKeys
    bs._requestRemoteConfigKeys(gCloudConfigKeys)

def writeConfig(force=False,storeRemote=True):
    global gConfigFileIsHealthy
    if not gConfigFileIsHealthy and not force:
        print "Current config file is broken; skipping write to avoid losing settings."
        return
    path = bs.getConfigFilePath()

    # write to a temp file then swap in.. this should give us a decently foolproof write
    # (though maybe we should sync before the rename?....)
    s = ''.join(_prettifyDict(bs.config,0))

    try:
        f=open(path+'.tmp','w')
        f.write(s)
        f.close()
    except Exception,e:
        bs.screenMessage("Error writing config file:\n"+str(e))
        raise e

    # on windows rename doesn't atomically overwrite existing files, so lets try to delete if one exists
    if 'windows' in bs.getUserAgentString():
        try: os.remove(path)
        except Exception: pass

    try:
        os.rename(path+'.tmp',path)
    except Exception,e:
        bs.screenMessage("Error writing config file:\n"+str(e))
        raise e

    gConfigFileIsHealthy = True

    if storeRemote:
        # we want to send certain config keys to our remote config (if available)
        global gCloudConfig
        global gCloudConfigKeys
        configToSend = {}
        for key in gCloudConfigKeys:
            if key in bs.config:
                # we use our prettify func here instead of repr when possible
                # because it sorts dicts; otherwise our repr() may change even though the value hasn't
                if type(bs.config[key]) == dict:
                    valStr = ''.join(_prettifyDict(bs.config[key],0,False))
                elif type(bs.config[key]) == list:
                    valStr = ''.join(_prettifyList(bs.config[key],0,False))
                else: valStr = repr(bs.config[key])

                if key not in gCloudConfig or valStr != gCloudConfig[key]:
                    gCloudConfig[key] = configToSend[key] = valStr

        if len(configToSend) > 0:
            # we *always* send launchCount; we use it as a way to prevent overwriting
            # of new data with old
            configToSend['launchCount'] = repr(bs.config['launchCount'])
            bs._storeRemoteConfig(configToSend)


def getDefaultFreeForAllPlaylist():
    return [
         {
            'level':'Doom Shroom',
            'settings':{
               'Epic Mode':False,
               'Kills to Win Per Player':10,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsDeathMatch.DeathMatchGame'
         },
         {
            'level':'Monkey Face',
            'settings':{
               'Chosen One Gets Gloves':True,
               'Chosen One Gets Shield':False,
               'Chosen One Time':30,
               'Epic Mode':0,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsChosenOne.ChosenOneGame'
         },
         {
            'level':'Zigzag',
            'settings':{
               'Hold Time':30,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsKingOfTheHill.KingOfTheHillGame'
         },
         {
            'level':'Tip Top',
            'settings':{
               'Epic Mode':1,
               'Lives Per Player':1,
               'Respawn Times':1.0,
               'Time Limit':120
            },
            'type':'bsElimination.EliminationGame'
         },
         {
            'level':'The Pad',
            'settings':{
               'Hold Time':30,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsKeepAway.KeepAwayGame'
         },
         {
            'level':'Rampage',
            'settings':{
               'Epic Mode':True,
               'Kills to Win Per Player':10,
               'Respawn Times':0.25,
               'Time Limit':120
            },
            'type':'bsDeathMatch.DeathMatchGame'
         },
         {
            'level':'Big G',
            'settings':{
               'Epic Mode':False,
               'Laps':3,
               'Mine Spawn Interval':4000,
               'Mine Spawning':4000,
               'Time Limit':300
            },
            'type':'bsRace.RaceGame'
         },
         {
            'level':'Happy Thoughts',
            'settings':{
               'Hold Time':30,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsKingOfTheHill.KingOfTheHillGame'
         },
         {
            'level':'Step Right Up',
            'settings':{
               'Epic Mode':False,
               'Lives Per Player':5,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsElimination.EliminationGame'
         },
         {
            'level':'Crag Castle',
            'settings':{
               'Epic Mode':False,
               'Kills to Win Per Player':10,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsDeathMatch.DeathMatchGame'
         }
      ]

def getDefaultTeamsPlaylist():
    return [
         {
            'level':'Bridgit',
            'settings':{
               'Epic Mode':False,
               'Flag Idle Return Time':30,
               'Flag Touch Return Time':0,
               'Respawn Times':1.0,
               'Score to Win':3,
               'Time Limit':600
            },
            'type':'bsCaptureTheFlag.CTFGame'
         },
         {
            'level':'Step Right Up',
            'settings':{
               'Epic Mode':False,
               'Respawn Times':1.0,
               'Score to Win':3,
               'Time Limit':600
            },
            'type':'bsAssault.AssaultGame'
         },
         {
            'level':'Rampage',
            'settings':{
               'Balance Total Lives':False,
               'Epic Mode':False,
               'Lives Per Player':3,
               'Respawn Times':1.0,
               'Solo Mode':True,
               'Time Limit':600
            },
            'type':'bsElimination.EliminationGame'
         },
         {
            'level':'Roundabout',
            'settings':{
               'Epic Mode':False,
               'Kills to Win Per Player':5,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsDeathMatch.DeathMatchGame'
         },
         {
            'level':'Hockey Stadium',
            'settings':{
               'Respawn Times':1.0,
               'Score to Win':1,
               'Time Limit':600
            },
            'type':'bsHockey.HockeyGame'
         },
         {
            'level':'Monkey Face',
            'settings':{
               'Hold Time':30,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsKeepAway.KeepAwayGame'
         },
         {
            'level':'Tip Top',
            'settings':{
               'Balance Total Lives':False,
               'Epic Mode':True,
               'Lives Per Player':1,
               'Respawn Times':1.0,
               'Solo Mode':False,
               'Time Limit':120
            },
            'type':'bsElimination.EliminationGame'
         },
         {
            'level':'Crag Castle',
            'settings':{
               'Epic Mode':False,
               'Respawn Times':1.0,
               'Score to Win':3,
               'Time Limit':300
            },
            'type':'bsAssault.AssaultGame'
         },
         {
            'level':'Doom Shroom',
            'settings':{
               'Epic Mode':False,
               'Kills to Win Per Player':5,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsDeathMatch.DeathMatchGame'
         },
         {
            'level':'Roundabout',
            'settings':{
               'Epic Mode':False,
               'Flag Idle Return Time':30,
               'Flag Touch Return Time':0,
               'Respawn Times':1.0,
               'Score to Win':2,
               'Time Limit':600
            },
            'type':'bsCaptureTheFlag.CTFGame'
         },
         {
            'level':'Football Stadium',
            'settings':{
               'Respawn Times':1.0,
               'Score to Win':21,
               'Time Limit':600
            },
            'type':'bsFootball.FootballTeamGame'
         },
         {
            'level':'Bridgit',
            'settings':{
               'Epic Mode':True,
               'Respawn Times':0.25,
               'Score to Win':3,
               'Time Limit':120
            },
            'type':'bsAssault.AssaultGame'
         },
         {
            'level':'Tip Top',
            'settings':{
               'Hold Time':30,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsKingOfTheHill.KingOfTheHillGame'
         },
         {
            'level':'Zigzag',
            'settings':{
               'Epic Mode':False,
               'Respawn Times':1.0,
               'Score to Win':2,
               'Time Limit':300
            },
            'type':'bsAssault.AssaultGame'
         },
         {
            'level':'Happy Thoughts',
            'settings':{
               'Epic Mode':False,
               'Flag Idle Return Time':30,
               'Flag Touch Return Time':0,
               'Respawn Times':1.0,
               'Score to Win':3,
               'Time Limit':300
            },
            'type':'bsCaptureTheFlag.CTFGame'
         },
         {
            'level':'Big G',
            'settings':{
               'Epic Mode':True,
               'Laps':1,
               'Mine Spawning':2000,
               'Time Limit':300
            },
            'type':'bsRace.RaceGame'
         },
         {
            'level':'Monkey Face',
            'settings':{
               'Epic Mode':False,
               'Kills to Win Per Player':5,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsDeathMatch.DeathMatchGame'
         },
         {
            'level':'Tip Top',
            'settings':{
               'Epic Mode':False,
               'Flag Idle Return Time':30,
               'Flag Touch Return Time':3,
               'Respawn Times':1.0,
               'Score to Win':2,
               'Time Limit':300
            },
            'type':'bsCaptureTheFlag.CTFGame'
         },
         {
            'level':'Crag Castle',
            'settings':{
               'Balance Total Lives':False,
               'Epic Mode':False,
               'Lives Per Player':3,
               'Respawn Times':1.0,
               'Solo Mode':False,
               'Time Limit':300
            },
            'type':'bsElimination.EliminationGame'
         },
         {
            'level':'Zigzag',
            'settings':{
               'Epic Mode':True,
               'Respawn Times':0.25,
               'Time Limit':120
            },
            'type':'bsConquest.ConquestGame'
         }
      ]

def filterPlaylist(playlist,sessionType,addResolvedType=False):
    """ returns a filtered version of a playlist - strips out or replaces invalid game types,
    makes sure all settings are present, and adds in a 'resolvedType' which is the actual type """
    
    goodList = []
    for i,gamespec in enumerate(playlist):

        # ok, for each game in our list, try to import the module and grab the actual game class.
        # add successful ones to our initial list to present to the user
        if type(gamespec['type']) in [str,unicode]:
            try:
                # do some type filters for backwards compat.
                if gamespec['type'] == 'Happy_Thoughts.HappyThoughtsGame': gamespec['type'] = 'bsAssault.AssaultGame'
                if gamespec['type'] == 'Assault.AssaultGame': gamespec['type'] = 'bsAssault.AssaultGame'
                if gamespec['type'] == 'King_of_the_Hill.KingOfTheHillGame': gamespec['type'] = 'bsKingOfTheHill.KingOfTheHillGame'
                if gamespec['type'] == 'Capture_the_Flag.CTFGame': gamespec['type'] = 'bsCaptureTheFlag.CTFGame'
                if gamespec['type'] == 'Death_Match.DeathMatchGame': gamespec['type'] = 'bsDeathMatch.DeathMatchGame'
                if gamespec['type'] == 'ChosenOne.ChosenOneGame': gamespec['type'] = 'bsChosenOne.ChosenOneGame'
                if gamespec['type'] == 'Conquest.Conquest': gamespec['type'] = 'bsConquest.ConquestGame'
                if gamespec['type'] == 'Conquest.ConquestGame': gamespec['type'] = 'bsConquest.ConquestGame'
                if gamespec['type'] == 'Elimination.EliminationGame': gamespec['type'] = 'bsElimination.EliminationGame'
                if gamespec['type'] == 'Football.FootballGame': gamespec['type'] = 'bsFootball.FootballTeamGame'
                if gamespec['type'] == 'Hockey.HockeyGame': gamespec['type'] = 'bsHockey.HockeyGame'
                if gamespec['type'] == 'Keep_Away.KeepAwayGame': gamespec['type'] = 'bsKeepAway.KeepAwayGame'
                if gamespec['type'] == 'Race.RaceGame': gamespec['type'] = 'bsRace.RaceGame'

                if gamespec['level'] == 'AlwaysLand': gamespec['level'] = 'Happy Thoughts'

                gameModuleName,gameClassName = gamespec['type'].split('.')

                gameModule = __import__(gameModuleName)
                gameClass = getattr(gameModule,gameClassName)

                if addResolvedType: gamespec['resolvedType'] = gameClass

                # make sure all settings the game defines are present
                neededSettings = gameClass.getSettings(sessionType)
                for settingName,setting in neededSettings:
                    if settingName not in gamespec['settings']:
                        gamespec['settings'][settingName] = setting['default']

                goodList.append(gamespec)

            except Exception,e:
                print 'Error resolving game \''+gamespec['type']+"':",e
        else:
            raise Exception("invalid gamespec format")

    return goodList


# used internally
def _handleRemoteAchievementList(completedAchievements):
    import bsAchievement
    bsAchievement.setCompletedAchievements(completedAchievements)

# used internally
def _handleRemoteConfigChange(config,overwrite):

    wroteSomething = False

    global gCloudConfig

    # if we receieved cloud config data that includes a launch-count substantially less than ours, lets
    # ignore the whole thing.  I've seen cases where initial cloud values never come in on
    # the first launch of an app, and this hopefully prevents that app from overwriting
    # everyone else's config when it spits its default values out to the cloud

    if overwrite and 'launchCount' in config and 'launchCount' in bs.config:
        # we listen to syncs that are witin one of our launch-count
        if config['launchCount'] < bs.config['launchCount'] - 1:
            return

    for c in config.items():

        try:
            # we store what we know the remote state to be to avoid sending redundant sets
            gCloudConfig[c[0]] = c[1]

            # eval the actual value from the string
            val = ast.literal_eval(c[1])

            # we commit the value to our local config only if overwrite is true or it doesn't exist
            # (and its a key we expect to find in the cloud)
            if c[0] in gCloudConfigKeys and (overwrite or (c[0] not in bs.config)):

                wroteSomething = True
                # do a bit of filtering here; for some values (such as, say, launchCount)
                # we want to use whatever's bigger and not just whichever's newer
                if c[0] == 'launchCount' and 'launchCount' in bs.config:
                    bs.config['launchCount'] = max(bs.config['launchCount'],val)

                # for any other vals we just overwrite with the newer cloud version
                else:
                    bs.config[c[0]] = val

        except Exception,e:
            print 'Exception setting remote config value for',c[0],'to',c[1],':',e

    # commit this stuff locally..
    if wroteSomething:
        writeConfig(storeRemote=False)

gLogUploadTimerStarted = False
gLogHaveNew = False

def _handleBecomingActive():
    # if there's music playing externally, make sure we aren't playing ours
    import bsUtils
    if bs.isOSPlayingMusic():
        bsUtils.playMusic(None)

# high level way to launch a challenge game locally
def _handleRunChallengeGame(game):
    import bsCoopGame
    import bsUI
    if game == '': raise Exception("empty game name")
    campaignName,levelName = game.split(':')
    campaign = bsCoopGame.getCampaign(campaignName)
    levels = campaign.getLevels()
    # if this campaign is sequential, make sure we've completed the one before this
    if campaign.isSequential():
        for i in range(len(levels)):
            if levels[i]['name'] == levelName:
                break
            if not levels[i]['Complete']:
                bsUI.LockedErrorWindow(bs.translate('coopLevelNames',levelName),bs.translate('coopLevelNames',levels[i]['name']))
                return False

    # if we're an expired demo, bring up the purchase ui for convenience
    if bs._getDemoTimeRemaining() == 0:
        bsUI.showDemoExpiredUI()
        return

    # ok we're good to go..
    bsUI.gCampaignSelection = {'campaign':campaignName,'level':levelName}
    def _fadeEnd():
        bs.hostSession('bsCoopGame.CoopSession')
    bs.fadeLocalScreen(False,time=250,endCall=_fadeEnd)
    return True

# called on debug log prints..
# when this happens, we can upload our log to the server
# after a short bit if desired.
def _handleLog():
    global gLogUploadTimerStarted
    global gLogHaveNew
    gLogHaveNew = True

    if not gLogUploadTimerStarted:
        def _putLog():
            import bsUtils
            if not bsUtils._gSuppressDebugReports:

                try: sessionName = str(bs.getSession())
                except Exception: sessionName = 'unavailable'
                try: activityName = str(bs.getSession().getActivity())
                except Exception: activityName = 'unavailable'

                info = {'log':bs._getLog(),
                        'version':bs.getVersion(),
                        'build':bs.getBuildNumber(),
                        'userAgentString':bs.getUserAgentString(),
                        'session':sessionName,
                        'activity':activityName,
                        'fatal':0,
                        'userRanCommands':bs.hasUserRunCommands(),
                        'time':bs.getRealTime(),
                        'userModded':bs.hasUserMods(),
                        }
                def response(data):
                    # a non-None response means success; lets
                    # take note that we don't need to report further
                    # log info this run
                    if data is not None:
                        global gLogHaveNew
                        gLogHaveNew = False
                        bs._markLogSent()
                bsUtils.serverPut('bsLog',info,response)
        gLogUploadTimerStarted = True

        # delay our log upload slightly in case other pertinant info gets printed
        # between now and then
        bs.realTimer(3000,_putLog)

        # after a while, allow another log-put
        def _reset():
            global gLogUploadTimerStarted
            gLogUploadTimerStarted = False
            if gLogHaveNew: _handleLog()
        if not bs._isLogFull():
            bs.realTimer(1000*60*10,_reset)

def _shouldSubmitDebugInfo():
    try:
        if bs.isTestBuild(): return True
        else: return bs.config['Submit Debug Info']
    except Exception: return True



class Session(object):
    """ Wrangles activities. One of these should be an argument to bs.hostSession().
    Responsible for transitioning between games, score-screens, etc
    and maintaining state between them (players, teams, score tallies, etc) """

    def __init__(self,teamNames=('Good Guys',),
                 teamColors=((0.6,0.2,1.0),),
                 useTeamColors=True,
                 maxPlayers=8):

        import bsPowerup
        import bsBomb
        import bsFlag
        import bsSpaz

        self._useTeams = (teamNames is not None)

        self.teams = {}
        self.players = {}
        self._nextTeamID = 0
        self._activity = None
        self._ending = False
        self._maxPlayers = maxPlayers
        self._joinSound = bs.getSound("dripity")
        self._playerLeftSound = bs.getSound("playerLeft")

        if self._useTeams:
            for i in range(len(teamColors)):
                team = self.teams[self._nextTeamID] = bs.Team(teamID=self._nextTeamID,
                                                              name=GameActivity.getTeamNameTranslated(teamNames[i]),
                                                              color=teamColors[i])
                self._nextTeamID += 1
                self.onTeamJoined(team)

        # pull on some stuff to get media loading (fixme shoudln't need this here)
        bsSpaz.getFactory()
        bsFlag.getFactory()
        bsPowerup.getFactory()
        bsBomb.getFactory()

        self._lobby = bsLobby.Lobby(teams=self.teams.values() if self._useTeams else None,
                                    useTeamColors=useTeamColors,
                                    delegate=self,vOffset=-150 if self.getSessionType() == 'co-op' else -30)
        self.scoreSet = bsScoreSet.ScoreSet(self._lobby)

    def onPlayerRequest(self,playerID):
        """ called when a new player wants to join - return True or False to accept/reject """
        # if we're ending, allow no new players
        if self._ending: return False

        if len(self.players)+len(self._lobby.getChoosers()) >= self._maxPlayers:
            bs.playSound(bs.getSound('error'))
            bs.screenMessage("Player limit of "+str(self._maxPlayers)+' reached; no joiners allowed',color=(0.8,0,0))
            return False
        bs.playSound(self._joinSound)
        self._lobby.addChooser(playerID)
        return True


    def onPlayerLeft(self,playerID,quiet=False):
        """ called when a previously-accepted player leaves """

        # remove them from the game rosters
        if not quiet: bs.playSound(self._playerLeftSound)
            
        if playerID in self.players:

            if not quiet: bs.screenMessage(bs.getPlayerName(playerID) + " left the game.")

            player = self.players[playerID]
            team = player.getTeam()

            # kill user data
            player.actor = None

            # remove him from his team roster
            del(team.players[playerID])
            del(self.players[playerID])

            # remove the player from the activity if its in there
            # (it may not be in the activity if we havnt begun yet)
            if playerID in self._activity.players:
                del self._activity.players[playerID]
                self._activity.onPlayerLeft(player)

            # if we're a non-team game, remove the team
            # (every player has their own team)
            if not self._useTeams:
                del self.teams[team.teamID]
                self.handleTeamLeft(team)
                # same as above; team may not be in activity..
                if team.teamID in self._activity.teams:
                    del self._activity.teams[team.teamID]
                    self._activity.handleTeamLeft(team)

        # ..and from the chooser
        self._lobby.removeChooser(playerID)

    def end(self):
        self.setActivity(EndSessionActivity(self))
        self._ending = True # prevents further activity-mucking

    def onTeamJoined(self,team):
        pass

    def handleTeamLeft(self,team):
        pass

    def handleMessage(self,m):
        """ generic message handling """
        if isinstance(m,bsLobby.PlayerReadyMessage):
            self.handlePlayerReady(m.chooser)

        elif isinstance(m,EndActivityMessage):
            # if the whole session is shutting down, dont' allow the users to touch anything
            if self._ending: return

            if m.activity is self._activity:

                # if the activity hasn't begun yet, just make note of what to do after we begin
                if not m.activity._hasBegun:
                    if not m.activity._shouldEndImmediately or m.force:
                        m.activity._shouldEndImmediately = True
                        m.activity._shouldEndImmediatelyResults = m.results
                        m.activity._shouldEndImmediatelyDelay = m.delay
                # the activity has begun; get ready to end it..
                else:
                    if (not m.activity._hasEnded) or m.force:
                        m.activity._hasEnded = True
                        self._activityEndTimer = bs.Timer(m.delay,bs.WeakCall(self.onActivityEnd,m.activity,m.results),timeType='real')
                        # also throw a check in there to make sure it's dead within a reasonable amount of time
                        bs.realTimer(m.delay+5000,bs.WeakCall(self.verifyActivityDeath,weakref.ref(m.activity)))

        elif isinstance(m,PlayerProfilesChangedMessage):
            self._lobby.reloadProfiles()
            for player in self.players.values():
                if player.actor is not None:
                    player.actor.updateForPlayer()

        else: Session.handleMessage(self,m)

    def setActivity(self,activity):
        """ flip to a particular state """
        if activity is self._activity or self._ending: return

        # this influences whether menus pause the action or just lets it run
        bs._setIsPlaying(activity._isPlayingActivity)
        if activity._isSlowMotion == True:
            bs.setGameSpeed(0.333)
            bs.setSoundPitch(0.4)
        elif activity._isSlowMotion == False:
            bs.setGameSpeed(1.0)
            bs.setSoundPitch(1.0)
        
        #activity._session = weakref.ref(self)
        activity.onTransitionIn()

        # we may want to explicitly keep the old activity alive while the new one transitions in
        t = activity._transitionTime
        if t > 0:
            # if we have a previous activity, create a timer that does nothing
            # but hold on to a reference to it until the new one is transitioned in
            def _doNothing(a): pass
            bs.gameTimer(t,bs.Call(_doNothing,self._activity))

        # if we have a current activity, let it transition out before
        # we begin the new one; otherwise begin the new one immediately
        if self._activity is not None:
            self._activity._transitioningOut = True
            self._nextActivity = activity
        else:
            bs.gameTimer(0,bs.WeakCall(activity._begin,self))

        # hmm should we not set this until the old one has transitioned out?...
        self._activity = activity

    def getActivity(self):
        return self._activity

    def getSessionType(self):
        """Will return 'teams','freeForAll', or 'co-op'"""
        raise Exception("this should be overridden by subclass")

    def verifyActivityDeath(self,activityRef):
        activity = activityRef()
        if activity is not None:
            print 'Activity is not dying when told to:',activity
            print 'Please make sure references get released or use weak references.'
            print 'The following objects are referencing the activity:'
            refs = list(gc.get_referrers(activity))
            i = 1
            def _val(obj):
                s = str(obj)
                if len(s) > 100: s = s[:90]+'...'+s[-10:]
                return s
            for ref in refs:
                if type(ref) is types.FrameType: continue
                print '     ref',i,':',type(ref),':',_val(ref)
                if type(ref) is list or type(ref) is tuple or type(ref) is dict:
                    refs2 = list(gc.get_referrers(ref))
                    j = 1
                    for ref2 in refs2:
                        if type(ref2) is types.FrameType: continue
                        print '        ref '+str(i)+'\'s ref',j,':',type(ref),':',_val(ref2)
                        j += 1
                i += 1

    def getLobby(self):
        return self._lobby

    def _runNextActivity(self):
        if self._nextActivity is not None:
            bs.gameTimer(0,bs.WeakCall(self._nextActivity._begin,self))
            self._nextActivity = None

    def _getPlayerIcon(self,player):
        return self._lobby.getPlayerIcon(player.playerID)

    def handlePlayerReady(self,chooser):
        """ called when a player has chosen a character and is ready to join """

        # during joins, wait till all choosers are ready
        # then create all players at once
        if self._activity._isJoiningActivity:
            if self._lobby.checkAllReady():
                for c in self._lobby.getChoosers():
                    player = self._newPlayer(c)
                self._lobby.removeAllChoosers()

                # get our next activity and launch it after a short delay
                self.onActivityEnd(self._activity,{})

            else:
                return
        # otherwise just create players instantly
        else:
            player = self._newPlayer(chooser)
            self._lobby.removeChooser(chooser.getPlayerID())


    def _newPlayer(self,chooser):
        """ called to create a player """

        # pass it to the current activity if its begun (otherwise it'll get passed when begin is called)
        passToActivity = (self._activity is not None and self._activity.hasBegun())

        # if we're a non-team game, each player gets their own team
        # (keeps mini-game coding simpler if we can always deal with teams)
        if self._useTeams:
            team = chooser.getTeam()
        else:
            ourTeamID = self._nextTeamID
            team = self.teams[ourTeamID] = bs.Team(teamID=ourTeamID,
                                                   name=bs.getPlayerName(chooser.getPlayerID()),
                                                   color=chooser.getColor())
            self._nextTeamID += 1
            self.onTeamJoined(team)

            if passToActivity:
                if ourTeamID in self._activity.teams: print "bs Warning fgwg: Duplicate team ID"
                self._activity.teams[ourTeamID] = team
                self._activity.onTeamJoined(team)

        newPlayer = bs.Player(playerID=chooser.getPlayerID(),
                              character=chooser.getCharacterName(),
                              team=team,
                              color=chooser.getColor(),
                              highlight=chooser.getHighlight(),
                              lobby=chooser.getLobby())
        self.players[chooser.getPlayerID()] = newPlayer
        self.scoreSet.registerPlayer(newPlayer)
        newPlayer.getTeam().players[chooser.getPlayerID()] = newPlayer

        if passToActivity:
            if chooser.getPlayerID() in self._activity.players: print "bs Warning f32f: Duplicate Player ID"
            self._activity.players[chooser.getPlayerID()] = newPlayer
            self._activity.onPlayerJoined(newPlayer)
        return newPlayer
        


class EndActivityMessage(object):
    def __init__(self,activity,results=None,delay=0,force=False):
        self.activity = activity
        self.results = results
        self.delay = delay
        self.force = force

class PlayerProfilesChangedMessage(object):
    """Local player profiles may have changed and should be reloaded if they are being used"""
    pass

class Activity(object):
    """ Units wrangled by a Session.  Encompasses things like games, score-screens, cutscenes, etc.
    There is generally 1 activity going on at any time, though they can overlap for transition purposes. """

    def __init__(self,session,settings={}):

        # the activity should preload as much of its media as it can
        # in its constructor, but not actually create any objects yet

        self.settings = settings
        self._session = weakref.ref(session)

        self._hasBegun = False
        self._shouldEndImmediately = False

        # joining activities are for waiting for initial player joins
        self._isJoiningActivity = False

        # influences the behavior of menus, etc (whether action gets paused)
        self._isPlayingActivity = False
        self._isSlowMotion = False

        # if the activity fades or transitions in, it should set the length of time here
        # so that previous activities will be kept alive for that long (avoiding holes in the screen)
        # also beware that some aspects of the old task won't get cleaned up until the end of this transition
        # so its best to avoid doing any real work until after your transition period is over and
        # your onBegin() is called
        self._transitionTime = 0

        # this gets set once another activity has begun transitioning in but before this one is killed.
        # make sure to not assign player inputs, change music, or anything else disruptive once this gets set.
        self._transitioningOut = False

        # a handy place to put most actors; this list is pruned of dead actors regularly
        # and these actors are insta-killed as the activity is dying.
        self._transientActors = []
        self._weakTransientActors = []
        self._ownedNodes = []
        self._ownedNodesCrossFade = []

        # this stuff gets filled in just before onBegin() is called
        self.teams = {}
        self.players = {}
        self.scoreSetbs = None

    def _pruneTransients(self):
        self._transientActors = [a for a in self._transientActors if a.exists()]
        self._weakTransientActors = [a for a in self._weakTransientActors if a() is not None and a().exists()]
        self._ownedNodes = [n for n in self._ownedNodes if bs.nodeExists(n)]
        self._ownedNodesCrossFade = [n for n in self._ownedNodesCrossFade if bs.nodeExists(n)]
        self._lastTransientPruneTime = bs.getGameTime()

    def _addTransientActor(self,a):
        if not isinstance(a,bs.Actor): raise Exception("non-actor passed to _addTransientActor")
        if self.hasBegun() and bs.getGameTime() - self._lastTransientPruneTime > 10000:
            print ('WARNING: it looks like transients are not being pruned in your activity;'
                   ' did you call Activity.onBegin() from your subclass?; '+str(self))
        self._transientActors.append(a)

    def _addWeakTransientActor(self,a):
        if not isinstance(a,bs.Actor): raise Exception("non-actor passed to _addWeakTransientActor")
        if self.hasBegun() and bs.getGameTime() - self._lastTransientPruneTime > 10000:
            print ('WARNING: it looks like transients are not being pruned in your activity;'
                   ' did you call Activity.onBegin() from your subclass?; '+str(self))
        self._weakTransientActors.append(weakref.ref(a))

    def _addOwnedNode(self,n,isCrossFade):
        if self.hasBegun() and bs.getGameTime() - self._lastTransientPruneTime > 10000:
            print ('WARNING: it looks like transients are not being pruned in your activity;'
                   ' did you call Activity.onBegin()? from your subclass; '+str(self))
        if isCrossFade:
            self._ownedNodesCrossFade.append(n)
        else:
            self._ownedNodes.append(n)
        
    def __del__(self):

        # do some default cleanup
        try:
            # send immediate die messages to our transient actors
            for actor in self._transientActors:
                actor.handleMessage(bs.DieMessage(immediate=True))

            # and any weak-reffed ones too
            for actorRef in self._weakTransientActors:
                actor = actorRef()
                if actor is not None: actor.handleMessage(bs.DieMessage(immediate=True))

            # send immediate die messages to any actors attached to players
            for player in self.players.values():
                if player.actor is not None:
                    player.actor.handleMessage(bs.DieMessage(immediate=True))
                player.actor = None
                player.data = {} # clear activity data

            # lastly delete remaining nodes
            for node in self._ownedNodes:
                bs.deleteNode(node,quiet=True)

            for team in self.teams.values():
                team.data = {} # clear activity data

        except Exception,e:
            print 'Exception during activity cleanup:',e
            pass

        # tell the session to fire up the next activity
        if self._transitioningOut:
            session = self._session()
            if session: session._runNextActivity()

    def getSession(self):
        if self._session is None: return None
        return self._session()

    # called for all new players *including* ones in the initial set
    def onPlayerJoined(self,player):
        pass

    # called just before a departing player gets pulled from player/team lists
    def onPlayerLeft(self,player):
        pass

    # called when a new team enters the activity (including the initial set)
    def onTeamJoined(self,team):
        pass

    # called when a team leaves the activity
    def handleTeamLeft(self,team):
        pass

    # start your activity transitioning in; fade in backgrounds, play sounds, etc
    # you don't have access to players or teams yet however (the previous activity still owns them)
    def onTransitionIn(self):
        pass

    # called once the previous activity has died (or timed out)
    # at this point your players, teams, and scoreSet values are filled in and
    # you're free to run your actual game stuff
    def onBegin(self):
        # start pruning our transient actors periodically
        self._lastTransientPruneTime = bs.getGameTime()
        self._transientPruneTimer = bs.Timer(5000,bs.WeakCall(self._pruneTransients),repeat=True)

    # misc message handling
    def handleMessage(self,m):
        pass
    
    def getPlayerIcon(self,player):
        return self._session()._getPlayerIcon(player)
        
    def end(self,results=None,delay=0,force=False):

        # if we had a standard time-limit that hadnt expired, stop it so it doesnt tick annoyingly
        if hasattr(self,'_standardTimeLimitTime') and self._standardTimeLimitTime > 0:
            self._standardTimeLimitTimer = self._standardTimeLimitText = None

        self.getSession().handleMessage(EndActivityMessage(self,results,delay,force))

    def isRunning(self):
        """ Returns true if onBegin has been called and end has not yet """
        return (self._hasBegun and not self._hasEnded)

    def hasBegun(self):
        return self._hasBegun
        
    def hasEnded(self):
        return self._hasEnded

    def isTransitioningOut(self):
        return self._transitioningOut

    def fadeToRed(self):
        cExisting = bs.getAttr(bs.globals,'tint')
        c = bs.newNode("combine",owner=self,attrs={'input0':cExisting[0],'input1':cExisting[1],'input2':cExisting[2],'size':3})
        bsUtils.animate(c,'input1',{0:cExisting[1],2000:0})
        bsUtils.animate(c,'input2',{0:cExisting[2],2000:0})
        bs.connectAttr(c,'output',bs.globals,'tint')

    # private call to set up onBegin
    def _begin(self,session):
        self.scoreSet = session.scoreSet

        # reset all player input
        for p in session.players: bs.resetPlayerInput(p)

        # we need to temporarily clear team player-lists so that players
        # won't be accessible there until they've been added to the activity
        oldPlayerLists = {}
        for teamID,team in session.teams.items():
            oldPlayerLists[teamID] = team.players
            team.players = {}

        # send player/team-joined messages one by one to fill this activity's lists
        for teamID,team in session.teams.items():
            if teamID in self.teams: raise Exception("Duplicate Team ID")
            if teamID != team.teamID: raise Exception("Corrupt Team Data")
            self.teams[teamID] = team
            self.onTeamJoined(team)

        # now add each player to the activity and to its team
        for playerID,player in session.players.items():
            self.players[playerID] = player
            player.getTeam().players[playerID] = player
            self.onPlayerJoined(player)

        self._hasBegun = True
        self._hasEnded = False
        self.onBegin()

        # if we've already been told to die, do so now
        if self._shouldEndImmediately: self.end(self._shouldEndImmediatelyResults,self._shouldEndImmediatelyDelay)

class EndSessionActivity(Activity):
    """ Special activity to fade out and end the current session """
    def __init__(self,session,settings={}):
        Activity.__init__(self,session,settings)
        self._transitionTime = 250 # keeps prev activity alive while we fade out

    def onTransitionIn(self):
        Activity.onTransitionIn(self)
        bs.fadeLocalScreen(False,time=250)
        bs.lockAllInput()

    def onBegin(self):
        Activity.onBegin(self)

        # ok, for now we want to be all anal-retentive when shutting down sessions
        # and try to ensure that all actors have a chance to clean up their nodes
        # so lets spin here until either that happens or we time out

        # any choosers we've got up will cause lingering nodes..
        self.getSession()._lobby.removeAllChoosers()

        self._checkNum = 0
        self._checkNodesTimer = bs.Timer(100,bs.WeakCall(self._checkNodes),repeat=True)

    def _checkNodes(self):
        nodesRemain = False
        for node in bs.getNodes():
            if not bsUtils._isValidRemainingNode(node):
                nodesRemain = True
                break
        self._checkNum += 1
        # after a second lets print some warnings
        # if nodesRemain: print len(bs.getNodes())
        if self._checkNum == 20:
            bsUtils.printLiveObjectWarnings('2 seconds into EndSessionActivity',ignoreSession=self.getSession(),ignoreActivity=self)
        if (not nodesRemain) or self._checkNum > 40:
            if nodesRemain: print 'Timed out on session shutdown waiting for nodes to die'
            bs.unlockAllInput()
            bs.hostSession('bsMainMenu.MainMenu','Main Menu',False)
            

class JoiningActivity(Activity):
    """ Standard joining activity that shows tips and waits for all players to join """
    def __init__(self,session,settings={}):
        Activity.__init__(self,session,settings)

        # this activity is a special 'joiner' activity
        # - it will get shut down as soon as all players have checked ready
        self._isJoiningActivity = True

    def __del__(self):
        Activity.__del__(self)

    def onTransitionIn(self):
        self._background = bsAnnounce.Background(self,bs.Actor.UNRETAINED,
                                                 fadeTime=500,startFaded=True,showLogo=True)
        self._tipsText = bsAnnounce.InfoText(self,bs.Actor.UNRETAINED)
        bsUtils.playMusic('CharSelect')

        self._joinInfo = self.getSession().getLobby().createJoinInfo()


class ScoreScreenActivity(Activity):
    """ A standard score screen that fades in and shows stuff for a while.
    After a specified delay, player input is assigned to send an EndActivityMessage. """

    def __init__(self,session,settings={}):
        Activity.__init__(self,session,settings)
        self._transitionTime = 500
        self._birthTime = bs.getGameTime()
        self._minViewTime = 5000

    def onPlayerJoined(self,player):
        Activity.onPlayerJoined(self,player)
        timeTillAssign = max(0,self._birthTime+self._minViewTime-bs.getGameTime())
        # if we're still kicking at the end of our assign-delay, assign this guy's input to trigger us
        bs.gameTimer(timeTillAssign,bs.WeakCall(self._safeAssign,player.playerID))

    def _safeAssign(self,playerID):
        # just to be extra careful, don't assign if we're transitioning out..
        # (though theoretically that would be ok)
        if not self.isTransitioningOut():
            bs.assignPlayerInput(playerID,('jumpPress','punchPress','bombPress','pickUpPress'),bs.WeakCall(self._playerPress))

    def _playerPress(self):
        if bs._getDemoTimeRemaining() == 0:
            import bsUI
            bsUI.showDemoExpiredUI()
            return
        self.end()

    def onTransitionIn(self,music='Scores',showTips=True):
        Activity.onTransitionIn(self)
        self._background = bsAnnounce.Background(self,bs.Actor.UNRETAINED,
                                                 fadeTime=500,startFaded=False,showLogo=True)
        if showTips:
            self._tipsText = bsAnnounce.InfoText(self,bs.Actor.UNRETAINED)
        bsUtils.playMusic(music)

    def onBegin(self,customContinueMessage=None):
        Activity.onBegin(self)
        # pop up a 'press any button to continue' statement after our min-view-time
        # show a 'press any button to continue..' thing after a bit..
        if bs.getInterfaceType() == 'desktop': s = bs.getResource('pressAnyKeyButtonText')
        else: s = bs.getResource('pressAnyButtonText')

        bsAnnounce.Text(self,bs.Actor.RETAINED,
                        customContinueMessage if customContinueMessage else s,
                        vAttach='bottom',
                        hAlign='center',
                        flash=True,
                        position=(0,10),
                        scale=0.8,
                        color=(0.5,0.7,0.5,0.5),
                        transition='inBottomSlow',
                        transitionDelay=self._minViewTime)


class GameActivity(Activity):
    
    tips = []

    @classmethod
    def configUI(cls,sessionType,config,editInfo,completionCall):
        """ class method to call for configuring this type of game.
        It is given an existing config (or None if a fresh one should be made)
        and it should call completionCall with a filled out config on success
        or None on cancel.
        Generally your class just needs to override getSettings()
        and getSupportedMaps() and can leave this alone """
        import bsUI
        bsUI.standardGameConfigUI(cls,sessionType,config,editInfo,completionCall)

    @classmethod
    def getName(c):
        try: return c.__module__.replace('_',' ')
        except Exception: return 'Untitled Game'

    @classmethod
    def getNameTranslated(c):
        name = c.getName()
        try: translated = bs.getResource('translations.gameNames')[name]
        except Exception: translated = None
        if translated is None: translated = name
        return translated

    @classmethod
    def getDescriptionTranslated(c,sessionType):
        description = c.getDescription(sessionType)
        try: translated = bs.getResource('translations.gameDescriptions')[description]
        except Exception: translated = None
        if translated is None: translated = description
        return translated

    @classmethod
    def getTeamNameTranslated(c,name):
        try: translated = bs.getResource('translations.teamNames')[name]
        except Exception: translated = None
        if translated is None: translated = name
        return translated

    @classmethod
    def getDescription(c,sessionType):
        return ''

    @classmethod
    def getSettings(self,sessionType):
        """ called by the default Game configUI;
        return a dict of config options that should be presented to the user """
        return {}

    @classmethod
    def getSupportedMaps(self,sessionType):
        """ called by the default Game configUI; should return the list
        of maps valid for this game-type """
        import bsMap
        return bsMap.getMapList("melee")

    @classmethod
    def getConfigDescription(self,config):
        """ called by the default Game configUI; given a config, return a short description for it """
        name = self.getNameTranslated()
        if 'Epic Mode' in config['settings'] and config['settings']['Epic Mode']:
            name = bs.getResource('epicNameFilterText').replace('${NAME}',name)
        s = name+" @ "+bsMap.getTranslatedMapName(bsMap.getFilteredMapName(config['level']))
        return s

    @classmethod
    def supportsSessionType(self,mode):
        """ given a mode (team, freeForAll, challenge, etc) returns whether the game supports it """
        #return True if mode == 'teams' else False
        return (mode in self.getSupportedSessionTypes())

    @classmethod
    def getSupportedSessionTypes(self):
        """ return supported game modes """
        return ['teams']


    def __init__(self,session,level,settings={}):
        Activity.__init__(self,session,settings)
        self._isPlayingActivity = True # allow pausing
        self._spawnSound = bs.getSound('spawn')
        self._showKillPoints = True # whether to show points for kills

        # go ahead and get our level loading
        self._levelType = bsMap.getMapClass(level)
        self._levelType.preload()

    def getInstanceName(self):
        return self.standardInstanceNameFilter(self.getNameTranslated())

    def getInstanceScoreBoardName(self):
        return self.getInstanceName()

    def getInstanceDescription(self):
        return self.standardInstanceDescriptionFilter(self.getDescription(self.getSession().getSessionType()))

    def getInstanceScoreBoardDescription(self):
        return ''

    def standardInstanceNameFilter(self,name):
        # add bit about epic mode
        return bs.getResource('epicNameFilterText').replace('${NAME}',name) if ('Epic Mode' in self.settings and self.settings['Epic Mode']) else name

    def standardInstanceDescriptionFilter(self,description):
        # add bit about epic mode
        return bs.getResource('epicDescriptionFilterText').replace('${DESCRIPTION}',description) if ('Epic Mode' in self.settings and self.settings['Epic Mode']) else description

    def onTransitionIn(self,music=None):
        Activity.onTransitionIn(self)

        # give our level a chance to override the music
        # (for happy-thoughts and other such themed maps)
        overrideMusic = self._levelType.getMusicType()
        if overrideMusic is not None: music = overrideMusic

        if music is not None:  bsUtils.playMusic(music)

    def onBegin(self,nameSound=None):
        Activity.onBegin(self)

        # we dont do this in onTransitionIn because it may depend on players/teams which arent available until now
        bs.gameTimer(1,bs.WeakCall(self.showScoreBoardInfo))
        bs.gameTimer(1000,bs.WeakCall(self.showInfo,nameSound))
        bs.gameTimer(2500,bs.WeakCall(self.showTip))

        # store some basic info about players present at start time
        self.initialPlayerInfo = [{'name':bs.getPlayerName(p.playerID),'character':p.character} for p in self.players.values()]

        # sort this by name so high score lists/etc will be consistent regardless of player join order..
        self.initialPlayerInfo.sort(key=lambda x:x['name'])


    def onPlayerJoined(self,player):
        self.spawnPlayer(player)

    def onPlayerLeft(self,player):
        player.actor = None

    def handleMessage(self,m):
        if isinstance(m,bsSpaz.PlayerSpazDeathMessage):
            # inform our score-set of the demise
            player = m.spaz.getPlayer()
            if player is None: return # player may have just left or something..

            killer = m.getKillerPlayer()
            self.scoreSet.playerLostSpaz(player,killed=m.killed,killer=killer)

            # award the killer points if he's on a different team
            if killer is not None and killer.getTeam() is not player.getTeam():
                pts,importance = m.spaz.getDeathPoints(m.how)
                if self.isRunning():
                    self.scoreSet.playerScored(killer,pts,kill=True,victimPlayerID=player.playerID,importance=importance,showPoints=self._showKillPoints)

        elif isinstance(m,bs.DieMessage):
            pass

    def showScoreBoardInfo(self):
        sbName = self.getInstanceScoreBoardName()

        # the description can be either a string or a sequence with args to swap in post-translation
        sbDesc = self.getInstanceScoreBoardDescription()
        if type(sbDesc) in [unicode,str]: sbDesc = [sbDesc] # handle simple string case
        if type(sbDesc[0]) not in [unicode,str]: raise Exception("Invalid format for instance description")
        try: translation = bs.getResource('translations.gameDescriptions')[sbDesc[0]]
        except Exception: translation = None
        if translation is None: translation = sbDesc[0]
        for i in range(len(sbDesc)-1):
            translation = translation.replace('${ARG'+str(i+1)+'}',str(sbDesc[i+1]))
        sbDesc = translation

        y = -37 if sbDesc == '' else -30
        self._gameScoreBoardNameText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                                    "text",{'text':sbName,
                                                            'translate':(20,y) if self.getSession().getSessionType() == 'freeForAll' else (45,y),
                                                            'hAttach':"left",
                                                            'vAttach':"top",
                                                            'color':(0.8,0.8,0.8,1.0),
                                                            'scale':[0.8]},
                                                    name='Game Scoreboard Name')
        bsUtils.animate(self._gameScoreBoardNameText.node,'opacity',{0:0.0,1000:1.0})

        self._gameScoreBoardDescriptionText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                                           "text",{'text':sbDesc,
                                                                   'translate':(22,-42) if self.getSession().getSessionType() == 'freeForAll' else (47,-42),
                                                                   'scale':[0.5],
                                                                   'hAttach':"left",
                                                                   'vAttach':"top",
                                                                   'color':(0.8,0.8,0.8,0.8)},
                                                           name='Game Scoreboard Desc')
        bsUtils.animate(self._gameScoreBoardDescriptionText.node,'opacity',{0:0.0,1000:1.0})

    def showInfo(self,nameSound):
        """ show the game description """
        bsAnnounce.ZoomText(self,bs.Actor.RETAINED,
                            self.getInstanceName(),lifespan=2500,jitter=2.0,position=(0,180),
                            flash=False,color=(0.93*1.25,0.9*1.25,1.0*1.25),trailColor=(0.15,0.05,1.0,0.0))
        bs.gameTimer(200,bs.WeakCall(bs.playSound,nameSound if nameSound is not None else bs.getSound('gong')))
            
        # the description can be either a string or a sequence with args to swap in post-translation
        desc = self.getInstanceDescription()
        if type(desc) in [unicode,str]: desc = [desc] # handle simple string case
        if type(desc[0]) not in [unicode,str]: raise Exception("Invalid format for instance description")
        try: translation = bs.getResource('translations.gameDescriptions')[desc[0]]
        except Exception: translation = None
        if translation is None: translation = desc[0]
        for i in range(len(desc)-1):
            translation = translation.replace('${ARG'+str(i+1)+'}',str(desc[i+1]))

        d = bs.newNode('text',
                       owner=self,
                       attrs={'vAttach':'center',
                              'hAttach':'center',
                              'hAlign':'center',
                              'color':(1,1,1,1),
                              'translate':(0,80),
                              'text':translation},name='Game Description')
        c = bs.newNode("combine",owner=d,attrs={'input0':1.0,'input1':1.0,'input2':1.0})
        bs.connectAttr(c,'output',d,'color')
        keys = {500:0,1000:1.0,2500:1.0,3000:0.0}
        bsUtils.animate(c,"input3",keys)
        bs.gameTimer(3000,bs.WeakCall(bs.deleteNode,d,quiet=True))


    def showTip(self):
        # if theres any tips left on the global list, display one..
        if len(self.tips) > 0:
            tip = self.tips.pop(random.randrange(len(self.tips)))
            tipTitle = bs.getResource('tipText')+':'
            icon = None
            sound = None
            if type(tip) == dict:
                if 'icon' in tip: icon = tip['icon']
                if 'sound' in tip: sound = tip['sound']
                tip = tip['tip']
            tip = bs.translate('tips',tip)
            baseTranslate = (50,50)
            tipScale = 0.8
            tipTitleScale = 1.2
            tipWidth = bs.getStringWidth(tip)*tipScale
            tipTitleWidth = bs.getStringWidth(tipTitle)*tipTitleScale

            t = bs.newNode('text',
                           owner=self,
                           attrs={'text':tip,
                                  'scale':[tipScale],
                                  'translate':(baseTranslate[0]-tipWidth*0.5,baseTranslate[1]),
                                  'hAlign':'left',
                                  'vAlign':'center',
                                  'vAttach':'bottom'},
                           name='Tip')
            t2 = bs.newNode('text',
                            owner=t,
                            attrs={'text':tipTitle,
                                   'scale':[tipTitleScale],
                                   'translate':(baseTranslate[0]-tipWidth*0.5-(20 if icon is None else 82),baseTranslate[1]+2),
                                   'hAlign':'right',
                                   'vAlign':'center',
                                   'vAttach':'bottom'},
                            name='Tip2')
            if icon is not None:
                img = bs.newNode('image',
                                 owner=self,
                                 attrs={'texture':icon,
                                        'translate':(baseTranslate[0]-tipWidth*0.5-40,baseTranslate[1]+1),
                                        'scale':(50,50),
                                        'opacity':1.0,
                                        'color':(1,1,1),
                                        'absoluteScale':True,
                                        'attach':'bottomCenter'})
                bsUtils.animate(img,'opacity',{0:0,1000:1,4000:1,5000:0})
                bs.gameTimer(5000,bs.WeakCall(bs.deleteNode,img))
            if sound is not None:
                bs.playSound(sound)

            c = bs.newNode("combine",owner=t,attrs={'input0':1.0,'input1':0.8,'input2':1.0})
            bs.connectAttr(c,'output',t,'color')
            bs.connectAttr(c,'output',t2,'color')
            bsUtils.animate(c,'input3',{0:0,1000:1,4000:1,5000:0})
            bs.gameTimer(5000,bs.WeakCall(bs.deleteNode,t))

    # uses a weak-ref so its safe to call in a timer or other event
    def spawnPlayerRef(self,playerRef):
        player = playerRef()
        if player is not None: self.spawnPlayer(player)

    def spawnPlayer(self,player):
        self.spawnSpazForPlayer(player)

    def respawnPlayer(self,player,respawnTime=None):
        if player is None: return

        if respawnTime is None:
            if len(player.getTeam().players) == 1: respawnTime = 3000
            elif len(player.getTeam().players) == 2: respawnTime = 5000
            elif len(player.getTeam().players) == 3: respawnTime = 6000
            else: respawnTime = 7000

        # if this standard setting is present, factor it in
        if 'Respawn Times' in self.settings: respawnTime *= self.settings['Respawn Times']

        respawnTime = int(max(1000,respawnTime))
        if respawnTime%1000 != 0: respawnTime -= respawnTime%1000 # we want whole seconds

        if player.actor and self.isRunning():
            player.data['respawnTimer'] = bs.Timer(respawnTime,bs.WeakCall(self.spawnPlayerRef,weakref.ref(player)))
            player.data['respawnIcon'] = bsSpaz.RespawnIcon(player,self,respawnTime)
        

    def spawnSpazForPlayer(self,player,position,angle=None):
        "Spawn and wire up a standard player spaz"

        name = bs.getPlayerName(player.playerID)
        color = player.color
        highlight = player.highlight
        lightColor = bsUtils.getNormalizedColor(color)
        displayColor = bsUtils.getSafeDisplayColor(color)
        player.actor = spaz = bsSpaz.PlayerSpaz(self,bs.Actor.UNRETAINED,
                                                color=color,
                                                highlight=highlight,
                                                character=player.character,
                                                player=player,
                                                hockey=self._level.isHockey,
                                                powerupsExpire=self.getSession().getSessionType() == 'co-op')
        bs.setAttr(spaz.node,'name',name)
        bs.setAttr(spaz.node,'nameColor',displayColor)
        spaz.connectToPlayer(player.playerID)
        self.scoreSet.playerGotNewSpaz(player,spaz)
        spaz.handleMessage(bs.StandMessage(position,angle if angle is not None else random.uniform(0,360)))
        t = bs.getGameTime()
        bs.playSound(self._spawnSound,1,position=(spaz.node,"translate"))
        light = bs.newNode("light",owner=self,attrs={"color":lightColor})
        bs.connectAttr(spaz.node,"translate",light,"translate")
        bsUtils.animate(light,"intensity",{0:0,250:1,500:0})
        bs.gameTimer(500,bs.WeakCall(bs.deleteNode,light))
        return spaz

    def celebrate(self,duration):
        for player in self.players.values():
            try: bs.nodeMessage(player.actor.node,'celebrate',duration)
            except Exception: pass

    def projectFlagStand(self,pos):
        # need to do this in a timer for it to work.. need to look into that.
        bs.gameTimer(1,bs.WeakCall(self._projectFlagStand,pos[:3]))
    def _projectFlagStand(self,pos):
        bs.emitBGDynamics(position=pos,emitType='flagStand')

    def _standardTimeLimitGetStr(self):
        value = self._standardTimeLimitTime
        return bsUtils.getTimeString(value*1000,centi=False)

    def setupStandardTimeLimit(self,duration):
        if duration <= 0: return
        self._standardTimeLimitTime = int(duration)
        self._standardTimeLimitTimer = bs.Timer(1000,bs.WeakCall(self._standardTimeLimitTick),repeat=True)
        self._standardTimeLimitText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                                   'text',{'vAttach':'top',
                                                           'hAttach':'center',
                                                           'hAlign':'left',
                                                           'color':(0.6,0.6,0.6,0.7),
                                                           'translate':(-25,-30),
                                                           'scale':[0.9],
                                                           'text':self._standardTimeLimitGetStr()})
    def _standardTimeLimitTick(self):
        self._standardTimeLimitTime -= 1
        if self._standardTimeLimitTime <= 10:
            if self._standardTimeLimitTime == 10:
                bs.setAttr(self._standardTimeLimitText.node,'scale',[1.3])
                bs.setAttr(self._standardTimeLimitText.node,'translate',(-30,-45))
                c = bs.newNode('combine',owner=self._standardTimeLimitText.node)
                bs.connectAttr(c,'output',self._standardTimeLimitText.node,'color')
                bsUtils.animate(c,"input0",{0:1, 150:1},loop=True)
                bsUtils.animate(c,"input1",{0:1, 150:0.5},loop=True)
                bsUtils.animate(c,"input2",{0:0.1, 150:0.0},loop=True)
                bs.setAttr(c,'input3',1.0)
            bs.playSound(bs.getSound('tick'))
        if self._standardTimeLimitTime <= 0:
            self._standardTimeLimitTimer = None
            self._endGame()
            n = bs.newNode('text',
                           owner=self,
                           attrs={'vAttach':'top','hAttach':'center',
                                  'hAlign':'center','color':(1,0.7,0,1),
                                  'translate':(0,-90),'scale':[1.2],'text':'Time Expired'})
            bs.playSound(bs.getSound('refWhistle'))
            bsUtils.animate(n,"scale",{0:0.0, 100:1.4, 150:1.2})

        bs.setAttr(self._standardTimeLimitText.node,'text',self._standardTimeLimitGetStr())

    def showZoomMessage(self,message,color,scale=0.8):
        # reserve a spot on the screen (in case we get multiple of these so they dont overlap)
        try: times = self._zoomMessageTimes
        except Exception: self._zoomMessageTimes = {}
        i = 0
        curTime = bs.getGameTime()
        duration = 2000
        while True:
            if i not in self._zoomMessageTimes or self._zoomMessageTimes[i] < curTime:
                self._zoomMessageTimes[i] = curTime + duration
                break
            i += 1
        bsAnnounce.ZoomText(self,bs.Actor.RETAINED,
                            message,lifespan=duration,jitter=4.0,position=(0,200-i*170*scale),scale=scale,
                            trail=False,color=color)

    def cameraFlash(self,duration=999):
        xSpread = 10
        ySpread = 5
        positions = [[-xSpread,-ySpread],[0,-ySpread],[0,ySpread],[xSpread,-ySpread],[xSpread,ySpread],[-xSpread,ySpread]]
        times = [0,2700,1000,1800,500,1400]

        # store this on our activity so we only have one at a time
        self._cameraFlash = []
        for i in range(6):
            light = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                 "light",{'translate':(positions[i][0],0,positions[i][1]),
                                          'radius':1.0,
                                          'lightVolumes':False,
                                          'heightAttenuated':False,
                                          'color':(0.2,0.2,0.8)})
            s = 1.87
            iScale = 1.3
            tcombine = bs.newNode("combine",owner=light.node,attrs={'input0':positions[i][0],'input1':0,'input2':positions[i][1]})
            bs.connectAttr(tcombine,'output',light.node,'translate')
            x = positions[i][0]
            y = positions[i][1]
            spd = 0.5 + random.random()
            spd2 = 0.5 + random.random()
            bsUtils.animate(tcombine,'input0',{0:x+0, 69*spd:x+10.0, 143*spd:x-10.0, 201*spd:x+0},loop=True)
            bsUtils.animate(tcombine,'input2',{0:y+0, 150*spd2:y+10.0, 287*spd2:y-10.0, 398*spd2:y+0},loop=True)
            
            bsUtils.animate(light.node,"intensity",{0:0, 20*s:0, 50*s:0.8*iScale, 80*s:0, 100*s:0},loop=True,offset=times[i])
            bs.gameTimer(int(times[i]+random.randint(1,duration)*40*s),bs.WeakCall(bs.deleteNode,light.node))
            self._cameraFlash.append(light)



import bsLobby
import bsScoreSet
import bsAnnounce
import bsSpaz
import bsMap
import bsUtils
